/********************************************************* {COPYRIGHT-TOP} ***
 * Licensed Materials - Property of IBM
 * 5725-Z22, 5725-Z63, 5725-U33, 5725-Z63
 *
 * (C) Copyright IBM Corporation 2016, 2020
 *
 * All Rights Reserved.
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 ********************************************************** {COPYRIGHT-END} **/
// Node module: apiconnect-assembly

'use strict';

angular.module('apiconnect-assembly').factory('ApimPolicyFactory', ['$q', function($q) {
  var policies = [];
  policies.push({
    info: {
      name: 'switch',
      categories: ['Logic'],
      display: {
        color: 'inherit',
        icon: 'call_split'
      },
      description: "Define sets of policies to execute for specific conditions. Multiple cases can be defined, and each case may be associated with an arbitrary condition. Only the first matching case will be executed."
    },
    classes: ["logic"],
    logic: true,
    gateways: ["micro-gateway", "datapower-gateway"],
    properties: {
      "type": "object",
      "additionalProperties": false,
      "required": ["case"],
      "properties": {
        "title": {
        "type": "string"
        },
        "description": {
        "type": "string"
        },
        "case": {
          "type": "array",
          "minItems": 1,
          "items": {
            "type": "object",
            "additionalProperties": false,
            "oneOf": [
              {"required": ["condition", "execute"]},
              {"required": ["otherwise"]}
            ],
            "properties": {
              "condition": {
                "type": "string"
              },
              "execute": {
                "type": "array"
              },
              "otherwise": {
                "type": "array"
              }
            }
          }
        }
      }
    }
  });
  policies.push({
    info: {
      name: 'if',
      categories: ['Logic'],
      display: {
        color: 'inherit',
        icon: 'call_made'
      }
    },
    classes: ["logic"],
    logic: true,
    gateways: ["micro-gateway", "datapower-gateway"],
    properties: {
      "type": "object",
      "additionalProperties": false,
      "required": ["condition"],
      "properties": {
        "title": {
          "type": "string"
        },
        "description": {
          "type": "string"
        },
        "condition": {
          "type": "string"
        },
        "execute": {
          type: "array"
        }
      }
    }
  });
  policies.push({
    info: {
      name: 'throw',
      categories: ['Logic'],
      display: {
        color: '#D0021B',
        icon: 'report_problem'
      }
    },
    gateways: ["micro-gateway", "datapower-gateway"],
    properties: {
      "type": "object",
      "additionalProperties": false,
      "required": ["name"],
      "properties": {
        "title": {
          "type": "string"
        },
        "name": {
          "type": "string"
        },
        "message": {
          "type": "string"
        }
      }
    }
  });

  var policyPartials = {
    "if": "src/html/policies/if.html",
    "throw": "src/html/policies/throw.html",
    "switch": "src/html/policies/switch.html"
  };

  return {
    getPolicies: function() {
      return $q(function(resolve) {
        var allPolicies = angular.copy(policies);
        resolve(allPolicies);
      });
    },
    getPolicyPartial: function(policyName) {
      var customPartial = policyPartials[policyName];
      return (customPartial) ? customPartial : "src/html/policies/generator.html";
    }
  };
}]);

